from .levels import command as levelsGet

def command(dataset, argsList):
    levels = levelsGet(dataset, argsList)
    sumLin = 0
    sumSq = 0
    for i in range(len(levels[0])-1):
        sumLin += levels[0][i+1] - levels[0][i]
        sumSq += (levels[0][i+1] - levels[0][i]) ** 2
    mean = sumLin / (len(levels[0])-1)
    # We know the whole population of the levels
    stddev = ((sumSq / (len(levels[0])-1)) - (mean ** 2)) ** 0.5
    return [[mean],[stddev]]



import matplotlib.pyplot as plt
from ..parser.parser import parseDatasetConfig
from ..parser.parser import writeParseFunctions 
from ..parser.parser import writeHeaderFunctions 
from ..parser.parser import writeFooterFunctions 
from ..parser.savepoint import parse as savepointParse 
import configparser
from ..parser.combine import commands
from ..parser.combine import runGroupData
from .fitter import plotFit 
from .transformer import transforms
from .decorator import decorations
import importlib
import os

# Search for default plot types
roots = []
modFilenames = []
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/plot_types"))
roots.append(root)
modFilenames.append(filenames)
# Search for custom plot types
if os.path.isdir(os.path.expanduser("~/.config/compphysutils/plot_types")):
    root, _, filenames = next(os.walk(os.path.expanduser("~/.config/compphysutils/plot_types")))
    roots.append(root)
    modFilenames.append(filenames)
# Import all plot types
plotTypes = {}
for i in range(len(roots)):
    for filename in modFilenames[i]:
        plotTypeName = filename.split(".")[0]
        if plotTypeName[0:2] == "__":
            # Skip __init__.py and similar commands
            continue
        spec = importlib.util.spec_from_file_location("compphysutils.graphics.plot_types."+plotTypeName, roots[i]+"/"+filename)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        plotTypes[plotTypeName] = mod.plot

class ColorIterator:
    def __init__(self, singleCycle="b"):
        self.singleCycle = singleCycle.split()
        self.currentIndex = 0
        self.cycleLen = len(self.singleCycle)

    def __iter__(self):
        return self

    def __next__(self):
        returnVal = self.singleCycle[self.currentIndex % self.cycleLen]
        self.currentIndex += 1
        return returnVal

class LinestyleIterator:
    def __init__(self, singleCycle="solid"):
        self.singleCycle = singleCycle.split()
        self.currentIndex = 0
        self.cycleLen = len(self.singleCycle)

    def __iter__(self):
        return self

    def __next__(self):
        returnVal = self.singleCycle[self.currentIndex % self.cycleLen]
        self.currentIndex += 1
        return returnVal

def plot(datasets, plotType="line", axes=False, **plotOptions):
    if not axes:
        axes = plt.gca()
    if plotType in plotTypes:
        plotTypes[plotType](datasets, axes, **plotOptions)
    else:
        # Defaults to scatter
        plotTypes["scatter"](datasets, axes, **plotOptions)
    axes.set_xlabel(plotOptions["xlabel"])
    axes.set_ylabel(plotOptions["ylabel"])
    if plotOptions["xlim"]:
        axes.set_xlim(*plotOptions["xlim"])
    if plotOptions["ylim"]:
        axes.set_ylim(*plotOptions["ylim"])
    # Insert tick labels
    if plotOptions["xticks"]:
        axes.set_xticks(plotOptions["xticks"][0], labels=plotOptions["xticks"][1])
    if plotOptions["yticks"]:
        axes.set_yticks(plotOptions["yticks"][0], labels=plotOptions["yticks"][1])
    return axes

def fromConfig(configFileName, axes=False, datasets={}):
    axesGiven = False
    if axes:
        axesGiven = True
    cfg = configparser.ConfigParser()
    cfg.read(configFileName)
    datasets.update(runGroupData(cfg, datasets, configFileName))
    # Now, run any transform commands
    if "transform" in cfg["plot"]:
        transformCommands = cfg["plot"].get("transform").split("\n")
        for commandLine in transformCommands:
            commandSplitLine = commandLine.split()
            commandName = commandSplitLine[0]
            datasets = transforms[commandName](datasets, commandSplitLine[1:])
    if "savepoint" in cfg["plot"]:
        savepointParse(cfg["plot"].get("savepoint"), "transform", datasets, writeParseFunctions, writeHeaderFunctions, writeFooterFunctions, "data_transform.out")
    # Now, datasets are complete, and we can read the plot group
    # Also include options that are set directly via type - should be reserved for options that are not usable for many plot types
    graphTypeSplit = cfg["plot"].get("type", "scatter").split()
    graphType = graphTypeSplit[0]
    colCoords = cfg.get("plot", "cols").split("\n")
    for i in range(len(colCoords)):
        colCoords[i] = colCoords[i].split()
    chosenDatasets = []
    for i in range(len(colCoords)):
        chosenDatasets.append([])
        for j in range(0,len(colCoords[i]),2):
            chosenDatasets[i].append(datasets[colCoords[i][j]][int(colCoords[i][j+1])])
    plotOptions = {}
    plotOptions["plotArgString"] = graphTypeSplit[1:]
    plotOptions["legend"] = cfg["plot"].getboolean("legend", True)
    plotOptions["legend-pos"] = cfg["plot"].get("legend-pos", "best")
    plotOptions["legend-cols"] = int(cfg["plot"].get("legend-cols", 1))
    if "xlim" in cfg["plot"]:
        plotOptions["xlim"] = list(map(float, cfg["plot"].get("xlim").split()))
    else:
        plotOptions["xlim"] = False
    if "ylim" in cfg["plot"]:
        plotOptions["ylim"] = list(map(float, cfg["plot"].get("ylim").split()))
    else:
        plotOptions["ylim"] = False
    if "font" in cfg["plot"]:
        plt.rcParams["font.family"] = cfg["plot"].get("font", "sans")
    if "font-size" in cfg["plot"]:
        plt.rcParams["font.size"] = int(cfg["plot"].get("font-size", 12))
    if "mathfont" in cfg["plot"]:
        plt.rcParams["mathtext.fontset"] = cfg["plot"].get("mathfont", "cm")
    plotOptions["colorCycle"] = cfg["plot"].get("colorCycle", "b")
    plotOptions["colorCycle"] = ColorIterator(plotOptions["colorCycle"])
    plotOptions["linestyleCycle"] = cfg["plot"].get("linestyleCycle", "solid")
    plotOptions["linestyleCycle"] = LinestyleIterator(plotOptions["linestyleCycle"])
    plotOptions["xlabel"] = cfg["plot"].get("xlabel", None)
    plotOptions["ylabel"] = cfg["plot"].get("ylabel", None)
    plotOptions["figfile"] = cfg["plot"].get("figfile", False)
    # Dataset labels
    plotOptions["datasetLabels"] = cfg["plot"].get("labels", False)
    if plotOptions["datasetLabels"]:
        plotOptions["datasetLabels"] = plotOptions["datasetLabels"].split("\n")
        if len(plotOptions["datasetLabels"]) < len(colCoords):
            toAdd = len(colCoords) - len(plotOptions["datasetLabels"])
            for i in range(toAdd):
                plotOptions["datasetLabels"].append(None)
    # Arguments supplied are the dataset name, convert it to a dataset that is then plotted
    for ticksName in ["xticks", "yticks"]:
        plotOptions[ticksName] = cfg["plot"].get(ticksName, False)
        if plotOptions[ticksName]:
            plotOptions[ticksName] = datasets[plotOptions[ticksName]]
    axes = plot(chosenDatasets, graphType, axes=axes, **plotOptions)
    # If fit is present, handle it
    fitIndex = 0
    if cfg["plot"].get("fit", False):
        fitLabels = cfg["plot"].get("fit-labels", False)
        numFits = len(cfg["plot"].get("fit").split("\n"))
        if not fitLabels:
            fitLabels = [False] * numFits
        else:
            fitLabels = fitLabels.split("\n")
        if len(fitLabels) < numFits:
            fitLabels += [False] * (numFits - len(fitLabels))
        prevFitParams = []
        fitColorIterator = ColorIterator(cfg["plot"].get("fit-colorCycle", "tab:blue tab:orange tab:green tab:cyan"))
        for allFitArgs in cfg["plot"].get("fit").split("\n"):
            fitArgs = allFitArgs.split()
            # TODO : Fit args?
            prevFitParams += list(plotFit(
                chosenDatasets[int(fitArgs[1])],
                fitArgs[0],
                axes,
                fitPoints=int(cfg["plot"].get("fit-points", 100)),
                fitLabel=fitLabels[fitIndex],
                showParams=cfg["plot"].getboolean("fit-show-params", True),
                fitColorCycle=fitColorIterator,
                paramsPlacement=cfg["plot"].get("params-placement", False),
                paramsOffset=len(prevFitParams),
                xMin=float(cfg["plot"].get("fit-xmin", False)),
                xMax=float(cfg["plot"].get("fit-xmax", False)),
                dirtyRun=cfg["plot"].getboolean("fit-dirty-run", False)
                ))
            fitIndex += 1
    # Handle decorations for main axes
    if cfg["plot"].get("decorate", False):
        decorationCommands = cfg["plot"].get("decorate").split("\n")
        for decorationArgs in decorationCommands:
            decorationSplit = decorationArgs.split()
            axes, datasets = decorations[decorationSplit[0]](axes, datasets, decorationSplit[1:])
    # Move the legend render here, even after decorations (which can also be annotated)
    if plotOptions["legend"]:
        axes.legend(loc=plotOptions["legend-pos"], ncol=plotOptions["legend-cols"])
    # If an inset directive is present, add an inset to the current axes
    if cfg["plot"].get("inset", False):
        insetArgs = cfg["plot"].get("inset").split()
        # Arguments are xpos, ypos, xwidth, ywidth
        insetAxes = axes.inset_axes(list(map(float, insetArgs[1:])))
        fromConfig(insetArgs[0], axes=insetAxes, datasets=datasets)
    # If axes are provided, assume figure is printed somewhere else
    # TODO : Is this a reasonable assumption?
    if axesGiven:
        return True
    if plotOptions["figfile"]:
        plt.savefig(plotOptions["figfile"], bbox_inches="tight")
    else:
        plt.show()



import argparse

unionAP = argparse.ArgumentParser()
unionAP.add_argument("new_name", help="New name of the dataset")
unionAP.add_argument("datasets_to_merge", nargs="+", help="Datasets, whose columns are to be added together, in the given order. First dataset determines the number of columns.")
def command(datasets, commandArgs):
    args = unionAP.parse_args(commandArgs)
    ncols = len(datasets[args.datasets_to_merge[0]])
    newcols = []
    for i in range(ncols):
        newcols.append([])
        for datasetName in args.datasets_to_merge:
            newcols[i] = newcols[i] + datasets[datasetName][i]
    datasets[args.new_name] = newcols
    return datasets



import configparser
from .VectorReal import VectorReal
from .VectorRepresentation import VectorRepresentation

# TODO : Need to decide about the location of default files containing possible configs and about dir scanning
def parseVectors(vectorsString, representation=False):
    """
    Returns the real/representation vectors based on the files given in the string
    """
    # If the strings are not present, return false
    if not vectorsString:
        return False
    vecs = []
    vecStrings = vectorsString.split()
    if not representation:
        for i in range(len(vecStrings)):
            vecs.append(VectorReal(*map(float, vecStrings[i].split(","))))
    else:
        for i in range(len(vecStrings)):
            vecs.append(VectorRepresentation(*map(int, vecStrings[i].split(","))))
    return vecs

def parseCharConfig(cfg):
    realBasis = parseVectors(cfg["vectors"]["real"])
    reprBasis = parseVectors(cfg["vectors"]["representation"], representation=True)
    planarBasis = parseVectors(cfg["vectors"].get("planar", False), representation=True)
    unitCellLength = float(cfg["base"]["unit"])
    elemName = cfg["base"]["atom"]
    return realBasis, reprBasis, unitCellLength, elemName, planarBasis

def readCrystalChar(filename):
    """
    Returns the representation and real basis of the crystal
    Real basis is still in units of unit cell length
    """
    cfg = configparser.ConfigParser()
    cfg.read(filename)
    return parseCharConfig(cfg)



def plot(datasets, axisObj, datasetLabels=False, **plotOptions):
    if not datasetLabels:
        datasetLabels = [None] * len(datasets)
    for dataIndex in range(len(datasets)):
        if len(datasets[dataIndex]) >= 4:
            # If at least four columns provided, plot errors on both axes
            axisObj.errorbar(datasets[dataIndex][0], datasets[dataIndex][2], xerr=datasets[dataIndex][1], yerr=datasets[dataIndex][3], capsize=4, lw=0, elinewidth=2, marker="p", ms=1, label=datasetLabels[dataIndex], color=next(plotOptions["colorCycle"]))
        else:
            # Otherwise, just plot yerr
            axisObj.errorbar(datasets[dataIndex][0], datasets[dataIndex][1], yerr=datasets[dataIndex][2], capsize=4, lw=0, elinewidth=2, marker="p", ms=2, label=datasetLabels[dataIndex], color=next(plotOptions["colorCycle"]))
    return axisObj



import re
import argparse

class EigerReference:
    # Object that ensures that only the converged eigenvalues are read
    # https://physics.nist.gov/cgi-bin/cuu/Value?hrev
    HtoEv = 27.211386245988

    def __init__(self, parserArgs="--unit eV"):
        self.reading = False
        self.readingLineRe = re.compile("\s*Nr\.\s*Orbital\s*Occupation\s*Energy")
        self.dataRe = re.compile("(\d+\.\d+)?\s*([\-\+]+\d+\.\d+)[H\s=]*([\-\+]+\d+\.\d+)")
        ap = argparse.ArgumentParser()
        ap.add_argument("--unit", help="Unit to be outputed", default="eV")
        self.args = ap.parse_args(parserArgs.split())
        self.energyUnit = self.args.unit

    def testLineReading(self, line):
        resultObj = self.readingLineRe.match(line)
        if resultObj:
            self.reading = True

    def continueReading(self, line):
        resultObj = self.dataRe.search(line)
        if not resultObj:
            self.reading = False
            return False
        energyH = resultObj.group(2)
        occupation = resultObj.group(1)
        if occupation == None:
            occupation = "0.0"
        if self.energyUnit == "eV":
            return [float(energyH)*self.HtoEv, float(occupation)]
        else:
            return [float(energyH), float(occupation)]

def line(line, ER):
    if not ER.reading:
        ER.testLineReading(line)
        return False
    # Reading
    result = ER.continueReading(line)
    if not result:
        # Stop reading
        return False
    # Real result present, convert to floats and return
    return result

def initParserObjects(parserArgs):
    return [EigerReference(parserArgs)]

argDefaults = "--unit eV"



from .Vector import Vector
import array

class VectorRepresentation(Vector):
    """
    Implements the vector representation by using fixed length array of int only
    """

    def __init__(self, *args):
        if len(args) < 3:
            self.components = array.array("i", [0,0,0])
        else:
            self.components = array.array("i", [0]*len(args))
        for i in range(len(args)):
            self.components[i] = args[i]

    def __eq__(self, other):
        """
        Importantly, vector representation implements the exact equals method, as this is necessary to eliminate numerical noise in float positions
        """
        if type(other) == VectorRepresentation:
            isEqual = True
            for i in range(len(self.components)):
                isEqual = isEqual and (self.components[i] == other.components[i])
            return isEqual
        else:
            raise TypeError("Equivalence relation unknown for VectorRepresentation and "+str(type(other)))

    def translateToBasis(self, basis):
        if len(basis) != len(self.components):
            #raise ArithmeticError("The basis size "+str(len(basis))+" and the vector size "+str(len(self.components))" are different!")
            raise ArithmeticError("Different basis and vector size!")
        return sum(map(lambda x: x[0] * x[1], zip(self.components, basis)), basis[0].cloneZeros())




import argparse

levelsGetAP = argparse.ArgumentParser()
levelsGetAP.add_argument("--above", default=True, action="store_true", help="Scan for the given number of levels above HOMO.")
levelsGetAP.add_argument("--below", default=False, action="store_true", help="Scan for the given number of levels below and including HOMO.")
levelsGetAP.add_argument("--fractional", default=False, action="store_true", help="Instead of integer occupation, set LUMO to first level where occupation is less than one.")
levelsGetAP.add_argument("--homo_align", default=False, action="store_true", help="Set HOMO level to 0, offseting the other energy levels accordingly.")
levelsGetAP.add_argument("levelNum", default=3, type=int, help="Number of levels to scan for.")
def command(dataset, argsList):
    # This is just a helper function that is used in specific commands
    args = levelsGetAP.parse_args(argsList)
    levels= []
    for i in range(args.levelNum):
        levels.append(0.0)
    if args.below:
        homoIndex = 0
        for i in range(len(dataset[0])):
            # Scan for HOMO
            if dataset[1][i] == 0 or (dataset[1][i] < 1.0 and args.fractional):
                homoIndex = i-1
                break
        for i in range(args.levelNum):
            levels[i] = dataset[0][homoIndex - args.levelNum + i + 1]
        if args.homo_align:
            for i in range(len(levels)):
                levels[i] = levels[i] - dataset[0][homoIndex]
    else:
        # Assume above
        lumoIndex = 0
        for i in range(len(dataset[0])):
            # Scan for LUMO
            if dataset[1][i] == 0 or (dataset[1][i] < 1.0 and args.fractional):
                lumoIndex = i
                break
        for i in range(args.levelNum):
            levels[i] = dataset[0][lumoIndex + i]
        if args.homo_align:
            for i in range(len(levels)):
                levels[i] = levels[i] - dataset[0][lumoIndex - 1]
    return [levels]



import importlib
import os
from .parser import parseDatasetConfig
from .parser import writeParseFunctions
from .parser import writeHeaderFunctions
from .parser import writeFooterFunctions
from .savepoint import parse as savepointParse

## Search for default combine commands
roots = []
modFilenames = []
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/combine_commands"))
roots.append(root)
modFilenames.append(filenames)
## Search for custom combine commands
if os.path.isdir(os.path.expanduser("~/.config/compphysutils/combine_commands")):
    root, _, filenames = next(os.walk(os.path.expanduser("~/.config/compphysutils/combine_commands")))
    roots.append(root)
    modFilenames.append(filenames)
## Import all commands
commands = {}
for i in range(len(roots)):
    for filename in modFilenames[i]:
        commandName = filename.split(".")[0]
        if commandName[0:2] == "__":
            # Skip __init__.py and similar commands
            continue
        spec = importlib.util.spec_from_file_location("compphysutils.parser.combine_commands."+commandName, roots[i]+"/"+filename)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        commands[commandName] = mod.command

def runGroupData(cfg, datasets, cfgFileName):
    if "data" in cfg:
        # First, load data from datasetfiles
        datasetfiles = cfg["data"].get("datasetfiles", False)
        if datasetfiles:
            for datasetFileName in datasetfiles.split():
                datasets.update(parseDatasetConfig(datasetFileName))
    # If there are in place defined datasets, they take priority
    datasets.update(parseDatasetConfig(cfgFileName))
    # If there is no data group, this is all that is done
    if not "data" in cfg:
        return datasets
    # Now, progress to the combine command
    if "combine" in cfg["data"]:
        combineCommands = cfg.get("data", "combine").split("\n")
        for commandLine in combineCommands:
            commandSplitLine = commandLine.split()
            commandName = commandSplitLine[0]
            datasets = commands[commandName](datasets, commandSplitLine[1:])
    # Finally, handle savepoint in the combine context
    if "savepoint" in cfg["data"]:
        savepointParse(cfg["data"].get("savepoint"), "combine", datasets, writeParseFunctions, writeHeaderFunctions, writeFooterFunctions, "data_combine.out")
    # End by returning the (changed) datasets
    return datasets



from .post_processor import postProcessCommands
from .savepoint import datasetParse as savepointDatasetParse
import configparser
import os
import importlib
import re

parserModules = {}
# Check for custom parser modules
customParsers = []
if os.path.isdir(os.path.expanduser("~/.config/compphysutils/parsers")):
    root, _, filenames = next(os.walk(os.path.expanduser("~/.config/compphysutils/parsers")))
    customParsers = list(map(lambda f: root+"/"+f, filenames))
# Check for default parser modules
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/parsers"))
defaultParsers = list(map(lambda f: root+"/"+f, filenames))
# Ideally, custom parsers should overwrite default parsers, not sure if this works
rexp = re.compile(".*/(.*)\.py")
for parserFileName in (defaultParsers + customParsers):
    parserType = rexp.search(parserFileName).group(1)
    if parserType[0:2] == "__":
        # Not a parser, but just a init or other module
        continue
    spec = importlib.util.spec_from_file_location("compphysutils.parser.parsers."+parserType, parserFileName)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    parserModules[parserType] = mod
lineParseFunctions = {}
writeParseFunctions = {}
writeHeaderFunctions = {}
writeFooterFunctions = {}
parserArgsDefaults = {}
initObjectsFunctions = {}
for parserName in parserModules:
    lineParseFunctions[parserName] = parserModules[parserName].line
    parserArgsDefaults[parserName] = parserModules[parserName].argDefaults
    initObjectsFunctions[parserName] = parserModules[parserName].initParserObjects
    if hasattr(parserModules[parserName], "writeLine"):
        writeParseFunctions[parserName] = parserModules[parserName].writeLine
    else:
        writeParseFunctions[parserName] = False
    if hasattr(parserModules[parserName], "writeHeaders"):
        writeHeaderFunctions[parserName] = parserModules[parserName].writeHeaders
    else:
        writeHeaderFunctions[parserName] = False
    if hasattr(parserModules[parserName], "writeFooters"):
        writeFooterFunctions[parserName] = parserModules[parserName].writeFooters
    else:
        writeFooterFunctions[parserName] = False

def parseFile(filename, filetype, parserArgs=False):
    if not parserArgs:
        parserArgs = parserArgsDefaults[filetype]
    file = open(filename, "r")
    datagroups = []
    parserObjects = initObjectsFunctions[filetype](parserArgs)
    currentParser = lineParseFunctions[filetype]
    for line in file:
        # Read line by line
        # Can return bool False if line is to be skipped
        readGroups = currentParser(line, *parserObjects)
        if readGroups:
            for i in range(len(readGroups)):
                if len(datagroups) > i:
                    datagroups[i].append(readGroups[i])
                else:
                    datagroups.append([readGroups[i]])
    file.close()
    return datagroups

def postProcess(datagroups, command, args):
   return postProcessCommands[command](datagroups, args)

def parseDatasetConfig(configFilename):
    cfg = configparser.ConfigParser()
    cfg.read(configFilename)
    datasets = {}
    for groupName in cfg.sections():
        if "dataset" in groupName:
            datasetName = groupName.split(".")[1]
            if "file" in cfg[groupName]:
                # Create datasets from file
                parserArgs = cfg[groupName].get("parser-args", False)
                datasets[datasetName] = parseFile(cfg[groupName]["file"], cfg[groupName]["filetype"], parserArgs=parserArgs)
            elif "list" in cfg[groupName]:
                # Create dataset from list, defaultly convert to float
                # TODO : Should there be som interface to different convertors?
                # A more sophisticated and decoupled list types
                splitList = cfg[groupName]["list"].split("\n")
                if "listtype" in cfg[groupName]:
                    if cfg[groupName]["listtype"] == "string":
                        datasets[datasetName] = list(map(lambda s: s.split(), splitList))
                    else:
                        datasets[datasetName] = list(map(lambda s: list(map(float, s.split())), splitList))
                else:
                    datasets[datasetName] = list(map(lambda s: list(map(float, s.split())), splitList))
            if "savepoint" in cfg[groupName]:
                savepointDatasetParse(cfg[groupName].get("savepoint"), "load", datasets[datasetName], writeParseFunctions, writeHeaderFunctions, writeFooterFunctions)
            if "post-process" in cfg[groupName]:
                commandSplit = cfg[groupName]["post-process"].split()
                if len(commandSplit) > 1:
                    datasets[datasetName] = postProcess(datasets[datasetName], commandSplit[0], commandSplit[1:])
                else:
                    datasets[datasetName] = postProcess(datasets[datasetName], commandSplit[0], [])
            if "savepoint" in cfg[groupName]:
                savepointDatasetParse(cfg[groupName].get("savepoint"), "post-process", datasets[datasetName], writeParseFunctions, writeHeaderFunctions, writeFooterFunctions, defFilename="data_post.out")
    return datasets



import random
import configparser
from .readCrystalChar import readCrystalChar, parseCharConfig
from .VectorRepresentation import VectorRepresentation
from .VectorReal import VectorReal
import os

def getUnitBasis(N):
    basis = []
    for i in range(N):
        basis.append(VectorRepresentation(*([0]*N)))
        basis[-1].components[i] = 1
    return basis

def checkPositionPresent(crystalRep, vecRep, basisRep):
    for otherRep in crystalRep:
        if vecRep.translateToBasis(basisRep) == otherRep.translateToBasis(basisRep):
            return True
    return False

def generateNextLevel(crystalRep, unitBasis, basisRep):
    """
    Grows the crystal by one layer.
    """
    # Iterate over positions of all atoms
    newLayer = []
    for vecRep in crystalRep:
        # Try adding each basis vector
        for unitRep in unitBasis:
            newVec = vecRep + unitRep
            if checkPositionPresent(crystalRep, newVec,basisRep) or checkPositionPresent(newLayer, newVec, basisRep):
                # Do not add this vector, skip it
                continue
            # Add the vector to the crystal
            newLayer.append(newVec)
    # The argument is augmented and returned
    return crystalRep + newLayer

def addAtoms(crystalRep, basisRep, planarBasis, addatoms, seed=False):
    if seed:
        random.seed(seed)
    for i in range(addatoms):
        # Choose atom at random
        atomIndex = random.randint(0,len(crystalRep)-1)
        newVec = crystalRep[atomIndex]
        # Create 1/-1 multiplier choices
        basisMultipliers = []
        for j in range(len(planarBasis)):
            basisMultipliers.append(random.choice([1,-1]))
        while checkPositionPresent(crystalRep, newVec, basisRep):
            basisIndex = random.randint(0, len(planarBasis)-1)
            newVec = newVec + (basisMultipliers[basisIndex] * planarBasis[basisIndex])
        crystalRep = crystalRep + [newVec]
    return crystalRep

def generateNLevels(N, unitBasis, basisRep):
    crystalRep = [VectorRepresentation(*([0]*len(unitBasis)))]
    for i in range(N-1):
        crystalRep = generateNextLevel(crystalRep, unitBasis, basisRep)
    return crystalRep

def getRealStructure(crystalRep, basisReal):
    crystal = []
    for vecRep in crystalRep:
        crystal.append(vecRep.translateToBasis(basisReal))
    return crystal

def getAtomData(crystal, unitLength, elemName):
    data = []
    for pos in crystal:
        data.append(list((unitLength * pos).components) + [elemName])
    return data

def generateCrystalRepresentation(basisRep, layers=1):
    unitBasis = getUnitBasis(len(basisRep))
    crystalRep = generateNLevels(layers, unitBasis, basisRep)
    # add-atoms
    #if not planarBasis:
    #    crystalRep = addAtoms(crystalRep, basisRep, unitBasis, addatoms, seed=seed)
    #else:
    #    crystalRep = addAtoms(crystalRep, basisRep, planarBasis, addatoms, seed=seed)
    return crystalRep, unitBasis

def generateCrystal(crystalCharFilename, layers=1, addatoms=0, seed=False):
    basisReal, basisRep, unitLength, elemName, planarBasis = readCrystalChar(crystalCharFilename)
    crystalRep = generateCrystalRepresentation(basisRep, planarBasis, layers=layers, addatoms=addatoms, seed=seed)
    crystal = getRealStructure(crystalRep, basisReal)
    # Finally, convert to standard format - atomic data format and multiply by unitLength
    return getAtomData(crystal, unitLength, elemName)

def moveAtom(crystalRep, basisRep, index, vector):
    newVec = crystalRep[index] + vector
    if checkPositionPresent(crystalRep, newVec, basisRep):
        raise ValueError("Atom would be moved to an already occupied crystal site!")
    else:
        crystalRep[index] = newVec
    return crystalRep

def addAtomAtPosition(crystalRep, basisRep, vector):
    if checkPositionPresent(crystalRep, vector, basisRep):
        raise ValueError("Atom would be added to an already occupied crystal site!")
    else:
        crystalRep.append(vector)
    return crystalRep

def delAtomAtPosition(crystalRep, basisRep, vector):
    # Delete if possible, otherwise just raise warning, not exception
    deleted = False
    for i in range(len(crystalRep)):
        if vector == crystalRep[i]:
            del crystalRep[i]
            deleted = True
            break
    if not deleted:
        print("No atom found at position "+str(vector))
    return crystalRep

def delAtomAtIndex(crystalRep, index):
    del crystalRep[index]
    return crystalRep

def getDataString(atomicData):
    string = str(len(atomicData))+"\n\n"
    for i in range(len(atomicData)):
        string = string + atomicData[i][-1].upper()+" "+" ".join(map(str, atomicData[i][:3])) + "\n"
    return string

def printData(atomicData):
    print(getDataString(atomicData))

def saveData(atomicData, filename):
    file = open(filename, "w")
    file.write(getDataString(atomicData))
    file.close()

def postProcessVectorSum(unitBasis, planarBasis, instructions):
    vectorSum = unitBasis[0].cloneZeros()
    for i in range(0,len(instructions),2):
        if instructions[i] == "planar":
            vectorSum = vectorSum + planarBasis[int(instructions[i+1])]
        elif instructions[i] == "basis":
            vectorSum = vectorSum + unitBasis[int(instructions[i+1])]
        elif instructions[i] == "mbasis":
            vectorSum = vectorSum - unitBasis[int(instructions[i+1])]
        elif instructions[i] == "mplanar":
            vectorSum = vectorSum - planarBasis[int(instructions[i+1])]
    return vectorSum

def createClusterFromConfig(configFilename):
    cfg = configparser.ConfigParser()
    cfg.set("DEFAULT", "indexdump", "False")
    cfg.read(configFilename)
    if not (("lattice" in cfg["cluster"]) or (("base" in cfg.sections()) and ("vectors" in cfg.sections()))):
        raise ValueError("Missing configuration - need a lattice configuration file [cluster]->lattice or [base] and [vectors] sections in the config file.")
    else:
        if "lattice" in cfg["cluster"]:
            # Check whether the lattice exists in cwd or in config directory, cwd takes priority
            if os.path.isfile(cfg["cluster"]["lattice"]):
                basisReal, basisRep, unitLength, elemName, planarBasis = readCrystalChar(cfg["cluster"]["lattice"])
            else:
                # Try config dir
                basisReal, basisRep, unitLength, elemName, planarBasis = readCrystalChar(os.path.expanduser("~/.config/compphysutils/lattices/")+cfg["cluster"]["lattice"])
        else:
            basisReal, basisRep, unitLength, elemName, planarBasis = parseCharConfig(cfg)
        crystalRep, unitBasis = generateCrystalRepresentation(basisRep, layers=int(cfg["cluster"]["layers"]), )
        # Remove the atoms if doing a sphere cut
        if "post-process" in cfg.sections():
            if "sphere-cut" in cfg["post-process"]:
                radius = float(cfg["post-process"].get("sphere-cut", 1.0))
                newCrystalRep = []
                for atom in crystalRep:
                    realPos = atom.translateToBasis(basisReal) * unitLength
                    if abs(realPos) <= radius:
                        newCrystalRep.append(atom)
                crystalRep = newCrystalRep
        # add-atoms
        if not planarBasis:
            crystalRep = addAtoms(crystalRep, basisRep, unitBasis, addatoms=int(cfg["cluster"]["addatoms"]), seed=cfg["cluster"].get("seed", False))
        else:
            crystalRep = addAtoms(crystalRep, basisRep, planarBasis, addatoms=int(cfg["cluster"]["addatoms"]), seed=cfg["cluster"].get("seed", False))
        # Start the post-process
        if "post-process" in cfg.sections():
            unitBasis = getUnitBasis(len(crystalRep[0].components))
            if "move" in cfg["post-process"]:
                for movementLine in cfg["post-process"]["move"].split("\n"):
                    movementInstructions = movementLine.split()
                    vectorSum = postProcessVectorSum(unitBasis, planarBasis, movementInstructions[1:])
                    # Final displacement vector complete - displace the atom
                    crystalRep = moveAtom(crystalRep, basisRep, int(movementInstructions[0]), vectorSum)
            # Movement instructions finished
            if "add" in cfg["post-process"]:
                for additionLine in cfg["post-process"]["add"].split("\n"):
                    additionInstructions = additionLine.split()
                    vectorSum = postProcessVectorSum(unitBasis, planarBasis, additionInstructions)
                    crystalRep = addAtomAtPosition(crystalRep, basisRep, vectorSum)
            if "delete" in cfg["post-process"]:
                for deletionLine in cfg["post-process"]["delete"].split("\n"):
                    deletionInstructions = deletionLine.split()
                    if deletionInstructions[0] == "index":
                        crystalRep = delAtomAtIndex(crystalRep, int(deletionInstructions[1]))
                    else:
                        vectorSum = postProcessVectorSum(unitBasis, planarBasis, deletionInstructions)
                        crystalRep = delAtomAtPosition(crystalRep, basisRep, vectorSum)
        structure = getRealStructure(crystalRep, basisReal)
        data = getAtomData(structure, unitLength, elemName)
        # If IndexDump is required, output atom data
        if "post-process" in cfg.sections():
            dumpIndexInfo = cfg.getboolean("post-process", "indexdump")
            if dumpIndexInfo:
                for i in range(len(crystalRep)):
                    print(i, crystalRep[i], data[i])
                if(planarBasis):
                    print("Planar basis")
                    for i in range(len(planarBasis)):
                        print(i, planarBasis[i])
        if "post-process" in  cfg.sections():
            if "random-noise" in cfg["post-process"]:
                # Generate random Gaussian noise with given amplitude for each atom in the system
                # This is applied at the very end in order to allow manipulation of crystal before
                argSplit = cfg["post-process"].get("random-noise").split()
                radius = float(argSplit[0])
                seed = int(argSplit[1])
                random.seed(seed)
                for i in range(len(data)):
                    for j in range(3):
                        data[i][j] += random.gauss(0,radius)
        saveData(data, cfg["cluster"]["name"])



import argparse

ap = argparse.ArgumentParser()
ap.add_argument("--linelength", default=1.0, type=float, help="Length of the line marking the level.")

def plot(datasets, axisObj, datasetLabels=False, **plotOptions):
    args = ap.parse_args(plotOptions["plotArgString"])
    if not datasetLabels:
        datasetLabels = [None] * len(datasets)
    for dataIndex in range(len(datasets)):
        # For x positions, only takes into account first element
        axisObj.eventplot(datasets[dataIndex][1], lineoffsets=datasets[dataIndex][0][0], linelengths=args.linelength, orientation="vertical", label=datasetLabels[dataIndex], colors=next(plotOptions["colorCycle"]))
    return axisObj



def plot(datasets, axisObj, datasetLabels=False, **plotOptions):
    if not datasetLabels:
        datasetLabels = [None] * len(datasets)
    for dataIndex in range(len(datasets)):
        axisObj.scatter(datasets[dataIndex][0], datasets[dataIndex][1], label=datasetLabels[dataIndex], c=next(plotOptions["colorCycle"]))
    return axisObj



class Vector:
    """
    Implements the basic setting and getting of components, scalar products, scalar multiplication and addition
    """

    def val_getter(self, valIndex):
        return self.components[valIndex]

    def val_setter(self, valIndex, val):
        self.components[valIndex] = val
        return self.components[valIndex]

    def val_deleter(self, valIndex):
        del self.components[valIndex]

    x = property(lambda self: self.val_getter(0),lambda self, val: self.val_setter(0, val), lambda self: self.val_deleter(0))
    y = property(lambda self: self.val_getter(1),lambda self, val: self.val_setter(1, val), lambda self: self.val_deleter(1))
    z = property(lambda self: self.val_getter(2),lambda self, val: self.val_setter(2, val), lambda self: self.val_deleter(2))

    def __init__(self, *args):
        if len(args) >= 3:
            self.components = [0]*len(args)
        else:
            self.components = [0,0,0]
        for i in range(len(args)):
            self.components[i] = args[i]

    def __str__(self):
        return "("+",".join(map(str, self.components))+")"

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        if issubclass(type(other), Vector):
            return type(self)(*map(lambda x: x[0]+x[1], zip(self.components, other.components)))
        else:
            raise TypeError("Addition not implemented for "+str(type(self))+" and "+str(type(other)))

    def __sub__(self, other):
        if issubclass(type(other), Vector):
            return type(self)(*map(lambda x: x[0]-x[1], zip(self.components, other.components)))
        else:
            raise TypeError("Subtraction not implemented for "+str(type(self))+" and "+str(type(other)))

    def __mul__(self, other):
        if issubclass(type(other), Vector):
            # Scalar (dot) product
            return sum(*map(lambda comps: comps[0]*comps[1], zip(self.components, other.components)))
        else:
            # Asume scalar multiplication
            res = type(self)(*map(lambda x: other * x, self.components))
            return res

    def __rmul__(self, other):
        # Multiplication is always symmetric
        return self.__mul__(other)

    def __neg__(self):
        # Only return scalar multiple
        return (-1) * self

    def __abs__(self):
        # Returns magnitude of the vector
        return (sum(map(lambda x:x**2, self.components)))**0.5

    def cloneZeros(self):
        return type(self)(*([0]*len(self.components)))



import argparse

averageAP = argparse.ArgumentParser()
averageAP.add_argument("--stderr", action="store_true", help="If given, second column of the new dataset includes standard error given by the average. Need at least two averaging datasets.")
averageAP.add_argument("new_name", help="Name of the dataset containing the averaged values.")
averageAP.add_argument("datasets_to_average", nargs="+", help="Datasets that are to be averaged, elementwise in each column. Number of columns is given by the first dataset.")

def command(datasets, commandArgs):
    args = averageAP.parse_args(commandArgs)
    averages = []
    stderrs = []
    ncols = len(datasets[args.datasets_to_average[0]])
    nrows = len(datasets[args.datasets_to_average[0]][0])
    N = len(args.datasets_to_average)
    for i in range(ncols):
        averages.append([])
        stderrs.append([])
    for rowIndex in range(nrows):
        for colIndex in range(ncols):
            sumLin = 0
            sumSq = 0
            for datasetName in args.datasets_to_average:
                sumLin += datasets[datasetName][colIndex][rowIndex]
                sumSq += datasets[datasetName][colIndex][rowIndex] ** 2
            averages[colIndex].append(sumLin / N)
            if args.stderr:
                stderrs[colIndex].append(((sumSq / N - averages[colIndex][rowIndex] ** 2)/(N-1)) ** 0.5)
    datasets[args.new_name] = []
    for i in range(ncols):
        datasets[args.new_name].append(averages[i])
        if args.stderr:
            datasets[args.new_name].append(stderrs[i])
    return datasets



from .Vector import Vector
import array

class VectorReal(Vector):
    """
    Implements the vector for real numbers, storing the components as fixed length array of floats
    """

    def __init__(self, *args):
        if len(args) < 3:
            self.components = array.array("f", [0]*3)
        else:
            self.components = array.array("f", [0]*len(args))
        for i in range(len(args)):
            self.components[i] = args[i]



def save(dataset, writeLine, filename, writeHeaders=False, writeFooters=False):
    nrows = len(dataset[0])
    ncols = len(dataset)
    file = open(filename, "w+")
    if writeHeaders:
        headers = writeHeaders(dataset)
        if len(headers) > 0:
            headers = headers + "\n"
        file.write(headers)
    for i in range(nrows):
        datarow = []
        for j in range(ncols):
            datarow.append(dataset[j][i])
        file.write(writeLine(datarow)+"\n")
    if writeFooters:
        footers = writeFooters(dataset)
        if len(footers) > 0:
            footers = footers + "\n"
        file.write(footers)
    file.close()

def datasetParse(savepointLines, context, dataset, writeLineFunctions, writeHeaderFunctions, writeFooterFunctions, defFilename="data.out"):
    # Parsers for the case when savepoint directive is provided at a given dataset - lacks one argument
    savepointSplit = savepointLines.split("\n")
    for i in range(len(savepointSplit)):
        savepointArgs = savepointSplit[i].split()
        if savepointArgs[0] == context:
            filename = defFilename
            if len(savepointArgs) > 2:
                filename = savepointArgs[2]
            # Save the dataset
            if not writeLineFunctions[savepointArgs[1]]:
                raise ValueError("No writeLine function defined for parser "+savepointArgs[1])
            save(dataset, writeLineFunctions[savepointArgs[1]], filename, writeHeaders=writeHeaderFunctions[savepointArgs[1]], writeFooters=writeFooterFunctions[savepointArgs[1]])

def parse(savepointLines, context, datasets, writeLineFunctions, writeHeaderFunctions, writeFooterFunctions, defFilename="data.out"):
    # Each line has following arguments - desired context, dataset name, parser with writeLine function, optional filename
    savepointSplit = savepointLines.split("\n")
    for i in range(len(savepointSplit)):
        savepointArgs = savepointSplit[i].split()
        if savepointArgs[0] == context:
            filename = defFilename
            if len(savepointArgs) > 3:
                filename = savepointArgs[3]
            # Save the dataset
            if not writeLineFunctions[savepointArgs[2]]:
                raise ValueError("No writeLine function defined for parser "+savepointArgs[2])
            save(datasets[savepointArgs[1]], writeLineFunctions[savepointArgs[2]], filename, writeHeaders=writeHeaderFunctions[savepointArgs[2]], writeFooters=writeFooterFunctions[savepointArgs[2]])
